@@ -43,14 +43,24 @@ Returns true if some file looks like a valid Unix password file
 =cut
 sub password_file
 {
-if (!$_[0]) { return 0; }
-elsif (&open_readfile(SHTEST, $_[0])) {
-	local($line);
-	$line = <SHTEST>;
-	close(SHTEST);
-	return $line =~ /^\S+:\S*:/;
+my ($file) = @_;
+if (!$file) {
+	return 0;
 	}
-else { return 0; }
+my $rv = $password_file_cache{$file};
+if (defined($rv)) {
+	return $rv;
+	}
+if (&open_readfile(SHTEST, $_[0])) {
+	my $line = <SHTEST>;
+	close(SHTEST);
+	$rv = $line =~ /^\S+:\S*:/ ? 1 : 0;
+	}
+else {
+	$rv = 0;
+	}
+$password_file_cache{$file} = $rv;
+return $rv;
 }
 
 =head2 list_users
@@ -123,7 +133,8 @@ if ($pft == 1) {
 	}
 elsif ($pft == 6) {
 	# Read netinfo dump
-	&open_execute_command(PASSWD, "nidump passwd '$netinfo_domain'", 1);
+	&open_execute_command(PASSWD,
+		"nidump passwd ".quotemeta($netinfo_domain), 1);
 	while(<PASSWD>) {
 		s/\r|\n//g;
 		if (/\S/ && !/^[#\+\-]/) {
@@ -141,7 +152,7 @@ elsif ($pft == 6) {
 elsif ($pft == 7) {
 	# Read directory services dump of users
 	&open_execute_command(PASSWD,
-		"dscl '$netinfo_domain' readall /Users", 1);
+		"dscl ".quotemeta($netinfo_domain)." readall /Users", 1);
 	local $user;
 	local $ls = $config{'lock_string'};
 	while(<PASSWD>) {
@@ -313,15 +324,22 @@ if ($pft == 1) {
 	}
 elsif ($pft == 3) {
 	# Just invoke the useradd command
-	&system_logged("useradd -u $_[0]->{'uid'} -g $_[0]->{'gid'} -c \"$_[0]->{'real'}\" -d $_[0]->{'home'} -s $_[0]->{'shell'} $_[0]->{'user'}");
+	&system_logged("useradd -u ".quotemeta($_[0]->{'uid'}).
+		" -g ".quotemeta($_[0]->{'gid'}).
+		" -c ".quotemeta($_[0]->{'real'}).
+		" -d ".quotemeta($_[0]->{'home'}).
+		" -s ".quotemeta($_[0]->{'shell'}).
+		" ".quotemeta($_[0]->{'user'}));
 	# And set the password
 	&system_logged("echo ".quotemeta($_[0]->{'pass'}).
 		       " | /usr/lib/scoadmin/account/password.tcl ".
-		       "$_[0]->{'user'} >/dev/null 2>&1");
+		       quotemeta($_[0]->{'user'})." >/dev/null 2>&1");
 	}
 elsif ($pft == 6) {
 	# Use the niutil command
-	&system_logged("niutil -create '$netinfo_domain' '/users/$_[0]->{'user'}'");
+	&system_logged("niutil -create ".
+		quotemeta($netinfo_domain)." ".
+		quotemeta("/users/$_[0]->{'user'}"));
 	&set_netinfo($_[0]);
 	}
 elsif ($pft == 7) {
@@ -428,15 +446,26 @@ if ($pft == 1) {
 	}
 elsif ($pft == 3) {
 	# Just use the usermod command
-	&system_logged("usermod -u $_[1]->{'uid'} -g $_[1]->{'gid'} -c \"$_[1]->{'real'}\" -d $_[1]->{'home'} -s $_[1]->{'shell'} $_[1]->{'user'}");
-	&system_logged("echo ".quotemeta($_[1]->{'pass'})." | /usr/lib/scoadmin/account/password.tcl $_[1]->{'user'}");
+	&system_logged("usermod -u ".quotemeta($_[1]->{'uid'}).
+		" -g ".quotemeta($_[1]->{'gid'}).
+		" -c ".quotemeta($_[1]->{'real'}).
+		" -d ".quotemeta($_[1]->{'home'}).
+		" -s ".quotemeta($_[1]->{'shell'}).
+		" ".quotemeta($_[1]->{'user'}));
+	&system_logged("echo ".quotemeta($_[1]->{'pass'}).
+		" | /usr/lib/scoadmin/account/password.tcl ".
+		quotemeta($_[1]->{'user'}));
 	}
 elsif ($pft == 6) {
 	# Just use the niutil command to update
 	if ($_[0]->{'user'} && $_[0]->{'user'} ne $_[1]->{'user'}) {
 		# Need to delete and re-create!
-		&system_logged("niutil -destroy '$netinfo_domain' '/users/$_[0]->{'user'}'");
-		&system_logged("niutil -create '$netinfo_domain' '/users/$_[1]->{'user'}'");
+		&system_logged("niutil -destroy ".
+			quotemeta($netinfo_domain)." ".
+			quotemeta("/users/$_[0]->{'user'}"));
+		&system_logged("niutil -create ".
+			quotemeta($netinfo_domain)." ".
+			quotemeta("/users/$_[1]->{'user'}"));
 		}
 	&set_netinfo($_[1]);
 	}
@@ -498,7 +527,12 @@ else {
 		# update AIX security user file as well..
 		# use chuser command because it's easier than working
 		# with the complexity issues of the file.
-		&system_logged("chuser expires=$_[1]->{'expire'} minage=$_[1]->{'min'} maxage=$_[1]->{'max'} pwdwarntime=$_[1]->{'warn'} $_[1]->{'user'}");
+		&system_logged("chuser expires=".
+			quotemeta($_[1]->{'expire'}).
+			" minage=".quotemeta($_[1]->{'min'}).
+			" maxage=".quotemeta($_[1]->{'max'}).
+			" pwdwarntime=".quotemeta($_[1]->{'warn'}).
+			" ".quotemeta($_[1]->{'user'}));
 		}
 	}
 if ($_[0] ne $_[1] && &indexof($_[0], @list_users_cache) != -1) {
@@ -536,15 +570,17 @@ if ($pft == 1) {
 	}
 elsif ($pft == 3) {
 	# Just invoke the userdel command
-	&system_logged("userdel -n0 $_[0]->{'user'}");
+	&system_logged("userdel -n0 ".quotemeta($_[0]->{'user'}));
 	}
 elsif ($pft == 4) {
 	# Just invoke the rmuser command
-	&system_logged("rmuser -p $_[0]->{'user'}");
+	&system_logged("rmuser -p ".quotemeta($_[0]->{'user'}));
 	}
 elsif ($pft == 6) {
 	# Just delete with the niutil command
-	&system_logged("niutil -destroy '$netinfo_domain' '/users/$_[0]->{'user'}'");
+	&system_logged("niutil -destroy ".
+		quotemeta($netinfo_domain)." ".
+		quotemeta("/users/$_[0]->{'user'}"));
 	}
 elsif ($pft == 7) {
 	# Delete from directory services
@@ -598,7 +634,8 @@ $lnum = 0;
 local $gft = &groupfiles_type();
 if ($gft == 5) {
 	# Get groups from netinfo
-	&open_execute_command(GROUP, "nidump group '$netinfo_domain'", 1);
+	&open_execute_command(GROUP,
+		"nidump group ".quotemeta($netinfo_domain), 1);
 	while(<GROUP>) {
 		s/\r|\n//g;
 		if (/\S/ && !/^[#\+\-]/) {
@@ -614,7 +651,7 @@ if ($gft == 5) {
 elsif ($gft == 7) {
 	# Read directory services dump of groups
 	&open_execute_command(PASSWD,
-		"dscl '$netinfo_domain' readall /Groups", 1);
+		"dscl ".quotemeta($netinfo_domain)." readall /Groups", 1);
 	local $group;
 	while(<PASSWD>) {
 		s/\r|\n//g;
@@ -727,7 +764,9 @@ sub create_group
 local $gft = &groupfiles_type();
 if ($gft == 5) {
 	# Use niutil command
-	&system_logged("niutil -create '$netinfo_domain' '/groups/$_[0]->{'group'}'");
+	&system_logged("niutil -create ".
+		quotemeta($netinfo_domain)." ".
+		quotemeta("/groups/$_[0]->{'group'}"));
 	&set_group_netinfo($_[0]);
 	}
 elsif ($gft == 7) {
@@ -785,8 +824,12 @@ if ($gft == 5) {
 	# Call niutil to update the group
 	if ($_[0]->{'group'} && $_[0]->{'group'} ne $_[1]->{'group'}) {
 		# Need to delete and re-create!
-		&system_logged("niutil -destroy '$netinfo_domain' '/groups/$_[0]->{'group'}'");
-		&system_logged("niutil -create '$netinfo_domain' '/groups/$_[1]->{'group'}'");
+		&system_logged("niutil -destroy ".
+			quotemeta($netinfo_domain)." ".
+			quotemeta("/groups/$_[0]->{'group'}"));
+		&system_logged("niutil -create ".
+			quotemeta($netinfo_domain)." ".
+			quotemeta("/groups/$_[1]->{'group'}"));
 		}
 	&set_group_netinfo($_[1]);
 	}
@@ -836,7 +879,9 @@ $_[0] || &error("Missing parameter to delete_group");
 local $gft = &groupfiles_type();
 if ($gft == 5) {
 	# Call niutil to delete
-	&system_logged("niutil -destroy '$netinfo_domain' '/groups/$_[0]->{'group'}'");
+	&system_logged("niutil -destroy ".
+		quotemeta($netinfo_domain)." ".
+		quotemeta("/groups/$_[0]->{'group'}"));
 	}
 elsif ($gft == 7) {
 	# Delete from directory services
@@ -1124,7 +1169,8 @@ else {
 	&set_ownership_permissions($_[2], $_[3], $st[2], "$_[1]/$base");
 	$nochown++;
 	}
-&system_logged("chown $opts -R $_[2]:$_[3] ".quotemeta("$_[1]/$base").
+&system_logged("chown $opts -R ".quotemeta($_[2]).":".
+	       quotemeta($_[3])." ".quotemeta("$_[1]/$base").
 	       " >/dev/null 2>/dev/null") if (!$nochown);
 return @rv;
 }
@@ -1436,15 +1482,26 @@ Update a NetInfo user based on a Webmin user hash. Mainly for internal use.
 sub set_netinfo
 {
 local %u = %{$_[0]};
-&system_logged("niutil -createprop '$netinfo_domain' '/users/$u{'user'}' passwd '$u{'pass'}'");
-&system_logged("niutil -createprop '$netinfo_domain' '/users/$u{'user'}' uid '$u{'uid'}'");
-&system_logged("niutil -createprop '$netinfo_domain' '/users/$u{'user'}' gid '$u{'gid'}'");
-&system_logged("niutil -createprop '$netinfo_domain' '/users/$u{'user'}' class '$u{'class'}'");
-&system_logged("niutil -createprop '$netinfo_domain' '/users/$u{'user'}' change '$u{'change'}'");
-&system_logged("niutil -createprop '$netinfo_domain' '/users/$u{'user'}' expire '$u{'expire'}'");
-&system_logged("niutil -createprop '$netinfo_domain' '/users/$u{'user'}' realname '$u{'real'}'");
-&system_logged("niutil -createprop '$netinfo_domain' '/users/$u{'user'}' home '$u{'home'}'");
-&system_logged("niutil -createprop '$netinfo_domain' '/users/$u{'user'}' shell '$u{'shell'}'");
+local $domain = quotemeta($netinfo_domain);
+local $upath = quotemeta("/users/$u{'user'}");
+&system_logged("niutil -createprop $domain $upath passwd ".
+	quotemeta($u{'pass'}));
+&system_logged("niutil -createprop $domain $upath uid ".
+	quotemeta($u{'uid'}));
+&system_logged("niutil -createprop $domain $upath gid ".
+	quotemeta($u{'gid'}));
+&system_logged("niutil -createprop $domain $upath class ".
+	quotemeta($u{'class'}));
+&system_logged("niutil -createprop $domain $upath change ".
+	quotemeta($u{'change'}));
+&system_logged("niutil -createprop $domain $upath expire ".
+	quotemeta($u{'expire'}));
+&system_logged("niutil -createprop $domain $upath realname ".
+	quotemeta($u{'real'}));
+&system_logged("niutil -createprop $domain $upath home ".
+	quotemeta($u{'home'}));
+&system_logged("niutil -createprop $domain $upath shell ".
+	quotemeta($u{'shell'}));
 }
 
 =head2 set_group_netinfo(&group)
@@ -1455,10 +1512,14 @@ Update a NetInfo group based on a Webmin group hash. Mainly for internal use.
 sub set_group_netinfo
 {
 local %g = %{$_[0]};
-local $mems = join(" ", map { "'$_'" } split(/,/, $g{'members'}));
-&system_logged("niutil -createprop '$netinfo_domain' '/groups/$g{'group'}' gid '$g{'gid'}'");
-&system_logged("niutil -createprop '$netinfo_domain' '/groups/$g{'group'}' passwd '$g{'pass'}'");
-&system_logged("niutil -createprop '$netinfo_domain' '/groups/$g{'group'}' users $mems");
+local $domain = quotemeta($netinfo_domain);
+local $gpath = quotemeta("/groups/$g{'group'}");
+local $mems = join(" ", map { quotemeta($_) } split(/,/, $g{'members'}));
+&system_logged("niutil -createprop $domain $gpath gid ".
+	quotemeta($g{'gid'}));
+&system_logged("niutil -createprop $domain $gpath passwd ".
+	quotemeta($g{'pass'}));
+&system_logged("niutil -createprop $domain $gpath users $mems");
 }
 
 =head2 set_user_dirinfo(&user)
@@ -1537,30 +1598,9 @@ if ($re && !eval { $pass =~ /^$re$/ }) {
 if ($config{'passwd_same'}) {
 	return &text('usave_epasswd_same') if ($pass =~ /\Q$username\E/i);
 	}
-if ($config{'passwd_dict'} && $pass =~ /^[A-Za-z\'\-]+$/ &&
-    (&has_command("ispell") || &has_command("spell"))) {
-	# Call spell or ispell to check for dictionary words
-	local $temp = &transname();
-	open(TEMP, ">$temp");
-	print TEMP $pass,"\n";
-	close(TEMP);
-	if (&has_command("ispell")) {
-		open(SPELL, "ispell -a <$temp |");
-		while(<SPELL>) {
-			if (/^(#|\&|\?)/) {
-				$unknown++;
-				}
-			}
-		close(SPELL);
-		}
-	else {
-		open(SPELL, "spell <$temp |");
-		local $line = <SPELL>;
-		$unknown++ if ($line);
-		close(SPELL);
-		}
-	unlink($temp);
-	return &text('usave_epasswd_dict') if (!$unknown);
+if ($config{'passwd_dict'} && $pass =~ /^[A-Za-z\'\-]+$/) {
+	# Check if dictionary word
+	return &text('usave_epasswd_dict') if (&is_dictionary_word($pass));
 	}
 if ($config{'passwd_prog'}) {
 	local $out;
@@ -1578,10 +1618,12 @@ if ($config{'passwd_prog'}) {
 		&print_tempfile(TEMP, $username,"\n");
 		&print_tempfile(TEMP, $pass,"\n");
 		&close_tempfile(TEMP);
-		$out = &backquote_command("$config{'passwd_prog'} <$temp 2>&1");
+		$out = &backquote_command(
+			"$config{'passwd_prog'} <".quotemeta($temp).
+			" 2>&1");
 		}
 	if ($?) {
-		return $out;
+		return $out || $text{'usave_epasswd_cmd'};
 		}
 	}
 if ($config{'passwd_mindays'} && $uinfo ne "none") {
@@ -1629,6 +1671,7 @@ if ($config{'max_length'} && length($username) > $config{'max_length'}) {
 local $re = $config{'username_re'};
 return &text('usave_ere', $re)
 	if ($re && !eval { $username =~ /^$re$/ });
+return $text{'usave_eltgt'} if ($username =~ /<|>/);
 return undef;
 }
 
@@ -1729,6 +1772,12 @@ if ($olduser) {
 	$ENV{'USERADMIN_OLD_GID'} = $olduser->{'gid'};
 	$ENV{'USERADMIN_OLD_PASS'} = $oldpass if (defined($oldpass));
 	}
+foreach my $f ("quota", "uquota", "mquota", "umquota") {
+	$ENV{'USERADMIN_'.uc($f)} = $user->{$f};
+	if ($olduser) {
+		$ENV{'USERADMIN_OLD_'.uc($f)} = $olduser->{$f};
+		}
+	}
 }
 
 =head2 set_group_envs(&group, action, [&oldgroup])
@@ -1772,7 +1821,7 @@ foreach $e (keys %ENV) {
 	}
 }
 
-=head2 encrypt_password(password, [salt])
+=head2 encrypt_password(password, [salt], [ignore-config-force-system-default], [no-sha1])
 
 Encrypts a password using the encryption format configured for this system.
 If the salt parameter is given, it will be used for hashing the password -
@@ -1783,9 +1832,15 @@ will be randomly generated.
 =cut
 sub encrypt_password
 {
-local ($pass, $salt) = @_;
+local ($pass, $salt, $force_system_default, $nosha1) = @_;
 local $format = 0;
-if ($gconfig{'os_type'} eq 'macos' && &passfiles_type() == 7) {
+my $format_error = sub {
+	my ($type, $format, $err) = @_;
+	&error(&text($type,
+	    "@{[&get_webprefix()]}/config.cgi?module=$module_name&section=line2",
+	    "@{[&get_webprefix()]}/cpan/download.cgi?source=3&cpan=$err", $err, $format));
+	};
+if (!$nosha1 && $gconfig{'os_type'} eq 'macos' && &passfiles_type() == 7) {
 	# New OSX directory service uses SHA1 for passwords!
 	$salt ||= chr(int(rand(26))+65).chr(int(rand(26))+65). 
 		  chr(int(rand(26))+65).chr(int(rand(26))+65);
@@ -1796,7 +1851,8 @@ if ($gconfig{'os_type'} eq 'macos' && &passfiles_type() == 7) {
 	elsif (&has_command("openssl")) {
 		# Use openssl command
 		local $temp = &transname();
-		&open_execute_command(OPENSSL, "openssl dgst -sha1 >$temp", 0);
+		&open_execute_command(OPENSSL,
+			"openssl dgst -sha1 >".quotemeta($temp), 0);
 		print OPENSSL $salt,$pass;
 		close(OPENSSL);
 		local $rv = &read_file_contents($temp);
@@ -1808,21 +1864,30 @@ if ($gconfig{'os_type'} eq 'macos' && &passfiles_type() == 7) {
 		&error("Either the Digest::SHA1 Perl module or openssl command is needed to hash passwords");
 		}
 	}
-elsif ($config{'md5'} == 2) {
+else {
+	my $config_md5 = $config{'md5'};
+	$config_md5 = 1 if ($force_system_default);
+
 	# Always use MD5
-	$format = 1;
-	}
-elsif ($config{'md5'} == 3) {
+	if ($config_md5 == 2) {
+		$format = 1;
+		}
 	# Always use blowfish
-	$format = 2;
-	}
-elsif ($config{'md5'} == 4) {
+	elsif ($config_md5 == 3) {
+		$format = 2;
+		}
 	# Always use SHA512
-	$format = 3;
-	}
-elsif ($config{'md5'} == 1 && !$config{'skip_md5'}) {
+	elsif ($config_md5 == 4) {
+		$format = 3;
+		}
+	# Always use yescrypt
+	elsif ($config_md5 == 5) {
+		$format = 4;
+		}
 	# Up to system
-	$format = &use_md5() if (defined(&use_md5));
+	elsif ($config_md5 == 1 && !$config{'skip_md5'}) {
+		$format = &use_md5() if (defined(&use_md5));
+		}
 	}
 
 if ($no_encrypt_password) {
@@ -1833,9 +1898,7 @@ elsif ($format == 1) {
 	# MD5 encryption is selected .. use it if possible
 	local $err = &check_md5();
 	if ($err) {
-		&error(&text('usave_edigestmd5',
-		    "/config.cgi?$module_name",
-		    "/cpan/download.cgi?source=3&cpan=$err", $err));
+		&$format_error('usave_edigestmod', 'MD5', $err);
 		}
 	return &encrypt_md5($pass, $salt);
 	}
@@ -1843,9 +1906,7 @@ elsif ($format == 2) {
 	# Blowfish is selected .. use it if possible
 	local $err = &check_blowfish();
 	if ($err) {
-		&error(&text('usave_edigestblowfish',
-		    "/config.cgi?$module_name",
-		    "/cpan/download.cgi?source=3&cpan=$err", $err));
+		&$format_error('usave_edigestmod', 'Blowfish', $err);
 		}
 	return &encrypt_blowfish($pass, $salt);
 	}
@@ -1853,10 +1914,18 @@ elsif ($format == 3) {
 	# SHA512 is selected .. use it
 	local $err = &check_sha512();
 	if ($err) {
-		&error($text{'usave_edigestsha512'});
+		&$format_error('usave_edigestcrypt', 'SHA512');
 		}
 	return &encrypt_sha512($pass, $salt);
 	}
+elsif ($format == 4) {
+	# yescrypt is selected .. use it
+	local $err = &check_yescrypt();
+	if ($err) {
+		&$format_error('usave_edigestcrypt', 'yescrypt');
+		}
+	return &encrypt_yescrypt($pass, $salt);
+	}
 else {
 	# Just do old-style crypt() DES encryption
 	if ($salt !~ /^[a-z0-9]{2}/i) {
@@ -2130,11 +2199,11 @@ sub mkuid
     foreach (split(//,$_[0])) {
       ++${num_let} if ( m/[a-z]/i );
     }
-    if ( length($_[0]) ne 7 ) {
+    if ( length($_[0]) != 7 ) {
         print "ERROR: Number of characters in username $_[0] is not equal to 7\n";
         return -1;
     }
-    if ( ${num_let} ne 3 && ${num_let} ne 4 ) {
+    if ( ${num_let} != 3 && ${num_let} != 4 ) {
         print "ERROR: Number of letters in username $_[0] is not equal to 3 or 4\n";
         return -1;
     }
@@ -2144,7 +2213,7 @@ sub mkuid
     my ${icnt} = -1;
     my ${lowuid};
     ${lowuid} = ( 26 ** ( ${num_let} - 1 ) * ${lowlimit}/100 ) + ${lowlimit};
-    ${lowuid} = ${lowlimit} if ( ${num_let} eq 3 );
+    ${lowuid} = ${lowlimit} if ( ${num_let} == 3 );
     my ${base} = 26;
 
 #################################################################
@@ -2309,6 +2378,15 @@ while(@ginfo = &my_getgrent()) {
 	$gidgrp{$ginfo[2]} = $ginfo[0];
 	}
 &my_endgrent();
+my $upagination = scalar(@{$users}) > $config{'display_max'};
+if ($upagination) {
+	my $upopts = \%in;
+	$upopts->{'paginations'}->{'offset'}->{'top'} = 155;
+	$upopts->{'paginations'}->{'offset'}->{'bottom'} = 170;
+	$upopts->{'paginations'}->{'form'} =
+	    { 'mode' => 'users', 'colspan' => 6 };
+	$upagination = &ui_paginations($users, $upopts);
+	}
 
 # Work out if any users can be edited
 local $anyedit;
@@ -2333,6 +2411,10 @@ if ($anyedit) {
 			&select_invert_link("d", $_[1]));
 	}
 push(@linksrow, @$links);
+if ($upagination) {
+	push(@{$rightlinks},
+	  "&nbsp;$upagination->{'search'}->{'form-data'}");
+	}
 local @grid = ( &ui_links_row(\@linksrow), &ui_links_row($rightlinks) );
 print &ui_grid_table(\@grid, 2, 100, [ "align=left", "align=right" ]);
 
@@ -2349,48 +2431,63 @@ print &ui_columns_start([
 	$lshow ? ( $text{'lastlogin'} ) : ( )
 	], 100, 0, \@tds);
 local $llogin;
+my $os_most_recent_logins = defined(&os_most_recent_logins);
 if ($lshow) {
-	$llogin = &get_recent_logins();
+	$llogin = &get_recent_logins() if ($os_most_recent_logins);
 	if (&foreign_check("mailboxes")) {
 		&foreign_require("mailboxes");
 		}
 	}
 local $u;
-foreach $u (@$users) {
-	$u->{'real'} =~ s/,.*$// if ($config{'extra_real'} ||
-				     $u->{'real'} =~ /,$/);
-	local @cols;
-	push(@cols, "") if ($anyedit && $u->{'noedit'});
-	push(@cols, &user_link($u));
-	push(@cols, $u->{'uid'});
-	push(@cols, &html_escape($gidgrp{$u->{'gid'}} || $u->{'gid'}));
-	push(@cols, &html_escape($u->{'real'}));
-	push(@cols, &html_escape($u->{'home'}));
-	push(@cols, &html_escape($u->{'shell'}));
-	if ($lshow) {
-		# Show last login, in local format after Unix time conversion
-		my $ll = $llogin->{$u->{'user'}};
-		if (defined(&mailboxes::parse_mail_date)) {
-			my $tm = &mailboxes::parse_mail_date($ll);
-			if ($tm) {
-				$ll = &make_date($tm);
+if (@$users) {
+	foreach $u (@$users) {
+		$u->{'real'} =~ s/,.*$// if ($config{'extra_real'} ||
+					     $u->{'real'} =~ /,$/);
+		local @cols;
+		push(@cols, "") if ($anyedit && $u->{'noedit'});
+		push(@cols, &user_link($u));
+		push(@cols, $u->{'uid'});
+		push(@cols, &html_escape($gidgrp{$u->{'gid'}} || $u->{'gid'}));
+		push(@cols, &html_escape($u->{'real'}));
+		push(@cols, &html_escape($u->{'home'}));
+		push(@cols, &html_escape($u->{'shell'}));
+		if ($lshow) {
+			# Show last login, in local format after Unix time conversion
+			$llogin = &get_recent_logins($u->{'user'}, 1)
+				if (!$os_most_recent_logins);
+			my $ll = $llogin->{$u->{'user'}};
+			if (defined(&mailboxes::parse_mail_date)) {
+				my $tm = &mailboxes::parse_mail_date($ll);
+				if ($tm) {
+					$ll = &make_date($tm);
+					}
 				}
+			push(@cols, &html_escape($ll));
+			}
+		if ($u->{'noedit'}) {
+			print &ui_columns_row(\@cols, \@tds);
+			}
+		else {
+			print &ui_checked_columns_row(\@cols, \@tds, "d", $u->{'user'});
 			}
-		push(@cols, &html_escape($ll));
-		}
-	if ($u->{'noedit'}) {
-		print &ui_columns_row(\@cols, \@tds);
-		}
-	else {
-		print &ui_checked_columns_row(\@cols, \@tds, "d", $u->{'user'});
 		}
 	}
+else {
+	print $upagination->{'search'}->{'no-results'};
+	}
 print &ui_columns_end();
+print $upagination->{'paginator'}->{'form-data'},
+      $upagination->{'paginator'}->{'form-scripts'}
+  if ($upagination);
 print &ui_links_row(\@linksrow);
 if ($anyedit) {
 	print $buttons;
 	print &ui_form_end();
 	}
+if ($upagination) {
+	print $upagination->{'paginator'}->{'form'};
+	print $upagination->{'search'}->{'form'};
+	}
 }
 
 =head2 groups_table(&groups, [form], [no-buttons], [&otherlinks], [&rightlinks])
@@ -2415,6 +2512,15 @@ foreach my $g (@$groups) {
 	}
 $anyedit = 0 if ($noboxes);
 
+my $gpagination = scalar(@{$groups}) > $config{'display_max'};
+if ($gpagination) {
+	my $gpopts = \%in;
+	$gpopts->{'paginations'}->{'offset'}->{'top'} = 155;
+	$gpopts->{'paginations'}->{'offset'}->{'bottom'} = 170;
+	$gpopts->{'paginations'}->{'form'} = { 'mode' => 'groups', 'colspan' => 4 };
+	$gpagination = &ui_paginations($groups, $gpopts);
+	}
+
 local @linksrow;
 if ($anyedit && $access{'gdelete'}) {
 	print &ui_form_start("mass_delete_group.cgi", "post");
@@ -2422,6 +2528,10 @@ if ($anyedit && $access{'gdelete'}) {
 			&select_invert_link("gd", $formno) );
 	}
 push(@linksrow, @$links);
+if ($gpagination) {
+	push(@{$rightlinks},
+	  "&nbsp;$gpagination->{'search'}->{'form-data'}");
+	}
 local @grid = ( &ui_links_row(\@linksrow), &ui_links_row($rightlinks) );
 print &ui_grid_table(\@grid, 2, 100, [ "align=left", "align=right" ]);
 
@@ -2434,33 +2544,45 @@ print &ui_columns_start([
 	$anydesc ? ( $text{'gedit_desc'} ) : ( ),
 	$text{'gedit_members'} ], 100, 0, \@tds);
 local $g;
-foreach $g (@$groups) {
-	local $members = join(" ", split(/,/, $g->{'members'}));
-	local @cols;
-	if ($anyedit && ($g->{'noedit'} || !$access{'gdelete'})) {
-		# Need an explicity blank first column
-		push(@cols, "");
-		}
-	push(@cols, &group_link($g));
-	push(@cols, $g->{'gid'});
-	if ($anydesc) {
-		push(@cols, &html_escape($g->{'desc'}));
-		}
-	push(@cols, &html_escape($members));
-	if ($g->{'noedit'} || !$access{'gdelete'}) {
-		print &ui_columns_row(\@cols, \@tds);
-		}
-	else {
-		print &ui_checked_columns_row(\@cols, \@tds, "gd",
-					      $g->{'group'});
+if (@$groups) {
+	foreach $g (@$groups) {
+		local $members = join(" ", split(/,/, $g->{'members'}));
+		local @cols;
+		if ($anyedit && ($g->{'noedit'} || !$access{'gdelete'})) {
+			# Need an explicitly blank first column
+			push(@cols, "");
+			}
+		push(@cols, &group_link($g));
+		push(@cols, $g->{'gid'});
+		if ($anydesc) {
+			push(@cols, &html_escape($g->{'desc'}));
+			}
+		push(@cols, &html_escape($members));
+		if ($g->{'noedit'} || !$access{'gdelete'}) {
+			print &ui_columns_row(\@cols, \@tds);
+			}
+		else {
+			print &ui_checked_columns_row(\@cols, \@tds, "gd",
+						      $g->{'group'});
+			}
 		}
 	}
+else {
+	print $gpagination->{'search'}->{'no-results'};
+	}
 print &ui_columns_end();
+print $gpagination->{'paginator'}->{'form-data'},
+      $gpagination->{'paginator'}->{'form-scripts'}
+  if ($gpagination);
 print &ui_links_row(\@linksrow);
 if ($anyedit && $access{'gdelete'}) {
 	print &ui_submit($text{'index_gmass'}, "delete"),"<br>\n";
 	print &ui_form_end();
 	}
+if ($gpagination) {
+	print $gpagination->{'paginator'}->{'form'};
+	print $gpagination->{'search'}->{'form'};
+	}
 }
 
 =head2 date_input(day, month, year, prefix)
@@ -2470,8 +2592,8 @@ Returns HTML for selecting a date
 =cut
 sub date_input
 {
-local ($d, $m, $y, $prefix) = @_;
-local $rv;
+my ($d, $m, $y, $prefix) = @_;
+my $rv;
 $rv .= &ui_textbox($prefix."d", $d, 3)."/";
 $rv .= &ui_select($prefix."m", $m,
 		[ map { [ $_, $text{"smonth_".$_} ] } (1..12) ])."/";
@@ -2487,11 +2609,13 @@ Returns a list of array references, each containing the details of a login.
 =cut
 sub list_last_logins
 {
-local @rv;
-&open_last_command(LAST, $_[0]);
+my ($user, $max) = @_;
+my @rv;
+undef($max) if (defined($max) && $max <= 0); # sanity check
+&open_last_command(LAST, $user, $max);
 while(@last = &read_last_line(LAST)) {
 	push(@rv, [ @last ]);
-	if ($_[1] && scalar(@rv) >= $_[1]) {
+	if ($max && scalar(@rv) >= $max) {
 		last;	# reached max
 		}
 	}
@@ -2499,19 +2623,20 @@ close(LAST);
 return @rv;
 }
 
-=head2 get_recent_logins()
+=head2 get_recent_logins([user], [max])
 
 Returns a hash ref from username to most recent login time/date
 
 =cut
 sub get_recent_logins
 {
+my ($user, $max) = @_;
 if (defined(&os_most_recent_logins)) {
 	return &os_most_recent_logins();
 	}
 else {
 	my %rv;
-	foreach my $l (&list_last_logins()) {
+	foreach my $l (&list_last_logins($user, $max)) {
 		$rv{$l->[0]} ||= $l->[3];
 		}
 	return \%rv;
@@ -2535,12 +2660,10 @@ if ($_[0]->{'noedit'}) {
 	return $dis;
 	}
 elsif ($_[0]->{'dn'}) {
-	return "<a href='edit_user.cgi?dn=".&urlize($_[0]->{'dn'})."'>".
-	       "$dis</a>";
+	return &ui_link("edit_user.cgi?dn=".&urlize($_[0]->{'dn'}), $dis);
 	}
 else {
-	return "<a href='edit_user.cgi?user=".&urlize($_[0]->{'user'})."'>".
-	       "$dis</a>";
+	return &ui_link("edit_user.cgi?user=".&urlize($_[0]->{'user'}), $dis);
 	}
 }
 
@@ -2555,12 +2678,10 @@ if ($_[0]->{'noedit'}) {
 	return &html_escape($_[0]->{'group'});
 	}
 elsif ($_[0]->{'dn'}) {
-	return "<a href='edit_group.cgi?dn=".&urlize($_[0]->{'dn'})."'>".
-	       &html_escape($_[0]->{'group'})."</a>";
+	return &ui_link("edit_group.cgi?dn=".&urlize($_[0]->{'dn'}), &html_escape($_[0]->{'group'}) );
 	}
 else {
-	return "<a href='edit_group.cgi?group=".&urlize($_[0]->{'group'})."'>".
-	       &html_escape($_[0]->{'group'})."</a>";
+	return &ui_link("edit_group.cgi?group=".&urlize($_[0]->{'group'}), &html_escape($_[0]->{'group'}) );
 	}
 }
 
@@ -2632,8 +2753,16 @@ $home ||= $user->{'home'};
 			   oct($config{'homedir_perms'}), $home) ||
 	&error(&text('usave_echmod', $!));
 if ($config{'selinux_con'} && &is_selinux_enabled() && &has_command("chcon")) {
-	&system_logged("chcon ".quotemeta($config{'selinux_con'}).
-		       " ".quotemeta($home)." >/dev/null 2>&1");
+	if ($config{'selinux_con'} eq "*") {
+		# Restore default context
+		&system_logged("restorecon -r ".
+			       quotemeta($home)." >/dev/null 2>&1");
+		}
+	else {
+		# Use specific context
+		&system_logged("chcon ".quotemeta($config{'selinux_con'}).
+			       " ".quotemeta($home)." >/dev/null 2>&1");
+		}
 	}
 &unlock_file($home);
 }
@@ -2650,8 +2779,12 @@ if ($user->{'home'} && -d $user->{'home'}) {
 	local $realhome = &resolve_links($user->{'home'});
 	local $qhome = quotemeta($realhome);
 	if ($config{'delete_only'}) {
-		&system_logged("find $qhome ! -type d -user $user->{'uid'} | xargs rm -f >/dev/null 2>&1");
-		&system_logged("find $qhome -type d -user $user->{'uid'} | xargs rmdir >/dev/null 2>&1");
+		&system_logged("find $qhome ! -type d -user ".
+			quotemeta($user->{'uid'}).
+			" | xargs rm -f >/dev/null 2>&1");
+		&system_logged("find $qhome -type d -user ".
+			quotemeta($user->{'uid'}).
+			" | xargs rmdir >/dev/null 2>&1");
 		&unlink_file($realhome);
 		}
 	else {
@@ -2705,4 +2838,18 @@ foreach (1 .. 15) {
 return $rv;
 }
 
+# can_user_sudo_root(user)
+# Returns 1 if some user can run all commands via sudo, 0 if not, or -1 if
+# sudo is not installed
+sub can_user_sudo_root
+{
+my ($user) = @_;
+my $sudo = &has_command("sudo");
+return -1 if (!$sudo);
+my $out = &backquote_command(
+	"$sudo -l -S -U ".quotemeta($user)." 2>&1 </dev/null");
+return -1 if ($?);
+return $out =~ /\(ALL\)\s+ALL|\(ALL\)\s+NOPASSWD:\s+ALL|\(ALL\s*:\s*ALL\)\s+ALL|\(ALL\s*:\s*ALL\)\s+NOPASSWD:\s+ALL/ ? 1 : 0;
+}
+
 1;
